import os
import configparser


def get_ansible_config_filepath() -> str:
    ansible_cfg_cur_dir_filepath = os.path.abspath("ansible.cfg")
    ansible_cfg_home_dir_filepath = os.path.expanduser("~/.ansible.cfg")
    ansible_etc_filepath = "/etc/ansible/ansible.cfg"
    if "ANSIBLE_CONFIG" in os.environ:
        ansible_config_env_filepath = os.environ["ANSIBLE_CONFIG"]
        if not os.path.exists(ansible_config_env_filepath):
            raise FileNotFoundError()
        return ansible_config_env_filepath
    elif os.path.exists(ansible_cfg_cur_dir_filepath):
        return ansible_cfg_cur_dir_filepath
    elif os.path.exists(ansible_cfg_home_dir_filepath):
        return ansible_cfg_home_dir_filepath
    elif os.path.exists(ansible_etc_filepath):
        return ansible_etc_filepath
    else:
        return ""


def get_dict_of_ansible_config_options() -> dict:
    config_filepath = get_ansible_config_filepath()
    if config_filepath == "":
        return {}
    config = configparser.ConfigParser()
    config.read(config_filepath)
    config_dictionary = {}
    for section in config.sections():
        for option in config.options(section):
            config_dictionary[option] = config.get(section, option)
    return config_dictionary



import logging
from typing import List

from ansible.inventory.host import Host  # type: ignore
from ansible.inventory.manager import InventoryManager  # type: ignore
from ansible.parsing.dataloader import DataLoader  # type: ignore

logger = logging.getLogger(__name__)


class InventoryAdapter:
    def __init__(self, inventory_path: str):
        self._inventory = InventoryManager(loader=DataLoader(), sources=inventory_path)

    def get_hosts_by_group(self, groups: List[str], no_groups: List[str]) -> List[Host]:
        output_hosts = set()  # type: ignore

        if not groups:
            output_hosts.update(self._inventory.hosts.values())
        else:
            for group in groups:
                group_data = self._inventory.groups[group]
                output_hosts.update(group_data.hosts)

        for no_group in no_groups:
            no_group_data = self._inventory.groups[no_group]
            output_hosts.difference_update(no_group_data.hosts)

        return list(output_hosts)

    def get_hosts_by_names(self, hostnames: List[str]) -> List[Host]:
        output_hosts = {self._inventory.hosts[hostname] for hostname in hostnames if
                        hostname in self._inventory.hosts}
        return list(output_hosts)

    def get_hosts_by_variables(self, hosts: List,
                               variables: List, no_variables: List) -> List[Host]:
        output_hosts = []
        if not no_variables:
            no_variables = []
        variable_husked_hosts = []
        if not variables:
            variable_husked_hosts = hosts
        else:
            for host in hosts:
                if any(var in host.get_vars().items()
                       for var in variables) or \
                        any(var[0] in host.get_vars().keys()
                            for var in variables if not var[1]):
                    variable_husked_hosts.append(host)
        for host in variable_husked_hosts:
            if not any(nvar in host.get_vars().items()
                       for nvar in no_variables) and \
                    not any(nvar[0] in host.get_vars().keys()
                            for nvar in no_variables if not nvar[1]):
                output_hosts.append(host)
        return output_hosts



import os
import datetime
from typing import List

from ansibleconnect.ansiblehostadapter import AnsibleHostAdapter


def in_tmux() -> bool:
    return 'TMUX' in os.environ


def tmux_start_command() -> str:
    tmux_session_or_window_name = datetime.datetime.now().strftime("ansibleconnect-%Y-%m-%d-%H-%M")
    if in_tmux():
        start_command = "tmux new-window -n {}".format(tmux_session_or_window_name)
    else:
        start_command = "tmux new-session -s {}".format(tmux_session_or_window_name)
    return start_command


def ssh_auth_socket_env_var_command() -> str:
    auth_socket = os.environ.get("SSH_AUTH_SOCK", "")
    if auth_socket != "":
        return "export SSH_AUTH_SOCK={}".format(auth_socket)
    else:
        return ""


def create_tmux_script(hosts: List[AnsibleHostAdapter]) -> str:
    tmux_file_lines = [tmux_start_command()]
    for index, host in enumerate(hosts):
        tmux_file_lines.append("send-keys '{extra_commands}; {conn_command}' C-m".format(
            extra_commands=ssh_auth_socket_env_var_command(),
            conn_command=host.connection_command))
        if index != len(hosts) - 1:
            tmux_file_lines.append("split-window -h")
            # tiled layout rearranges panes so that each pane has the same size
            tmux_file_lines.append("select-layout tiled")
    # \; is printed so that ; can be interpreted by tmux instead of shell
    tmux_script = " \\; ".join(tmux_file_lines)
    return tmux_script



ANSIBLE_NULL_VALUE = 'null'


def get_first_from_list_or_default(dictionary: dict, key_list: list, default_val=None):
    for key in key_list:
        if key in dictionary:
            return dictionary[key]
    return default_val


class ConnectionCommand:
    def __init__(self, host_name, host_variables: dict):
        self.host_name = host_name
        self.host = host_variables.get('ansible_host', None)
        self.user = host_variables.get('ansible_user', 'root')
        self.port = host_variables.get('ansible_port', 22)
        self.password = host_variables.get('ansible_password', None)


class SSHConnectionCommand(ConnectionCommand):
    # Order in these key lists is important as the first found in the dict will be returned
    # Variable names from https://docs.ansible.com/ansible/latest/plugins/connection/ssh.html
    # and https://docs.ansible.com/ansible/latest/reference_appendices/config.html
    SSH_HOST_KEYS = ['ansible_ssh_host', 'ansible_host']
    SSH_HOST_KEY_CHECKING_KEYS = ['ansible_ssh_host_key_checking', 'ansible_host_key_checking',
                                  'host_key_checking']
    SSH_PASSWORD_KEYS = ['ansible_ssh_password', 'ansible_ssh_pass', 'ansible_password']
    SSH_PORT_KEYS = ['ansible_ssh_port', 'ansible_port', 'remote_port']
    SSH_PRIVATE_KEY_FILE_KEYS = ['ansible_ssh_private_key_file', 'ansible_private_key_file',
                                 'private_key_file']
    SSH_USER_KEYS = ['ansible_ssh_user', 'ansible_user', 'remote_user']
    SSH_ARGS_KEYS = ['ansible_ssh_args', 'ssh_args']
    SSH_EXECUTABLE_KEYS = ['ansible_ssh_executable', 'ssh_executable']

    def __init__(self, host_name, host_variables: dict):
        super().__init__(host_name, host_variables)
        self.host = get_first_from_list_or_default(host_variables, self.SSH_HOST_KEYS, None)
        self.host_key_checking = get_first_from_list_or_default(host_variables,
                                                                self.SSH_HOST_KEY_CHECKING_KEYS,
                                                                True)
        self.password = get_first_from_list_or_default(host_variables, self.SSH_PASSWORD_KEYS, None)
        self.port = get_first_from_list_or_default(host_variables, self.SSH_PORT_KEYS, None)
        self.ssh_private_key_file = get_first_from_list_or_default(host_variables,
                                                                   self.SSH_PRIVATE_KEY_FILE_KEYS,
                                                                   None)
        self.user = get_first_from_list_or_default(host_variables, self.SSH_USER_KEYS, 'root')
        self.ssh_args = get_first_from_list_or_default(host_variables, self.SSH_ARGS_KEYS,
                                                       '-C -o ControlMaster=auto'
                                                       '-o ControlPersist=60s')
        self.ssh_common_args = host_variables.get('ansible_ssh_common_args', '')
        self.ssh_extra_args = host_variables.get('ansible_ssh_extra_args', '')
        self.ssh_executable = get_first_from_list_or_default(host_variables,
                                                             self.SSH_EXECUTABLE_KEYS, 'ssh')

    def _get_user_and_hostname(self):
        if self.host:
            user_and_hostname = '{user}@{host}'.format(user=self.user, host=self.host)
        else:
            # This case is useful when ssh connection information is held in
            # config file like ~/.ssh/config instead of the inventory file
            user_and_hostname = self.host_name
        return user_and_hostname

    def _get_ssh_options(self):
        ssh_options = ' '.join([self.ssh_args, self.ssh_extra_args, self.ssh_common_args])
        if self.ssh_private_key_file:
            ssh_options += ' -i {}'.format(self.ssh_private_key_file)
        if not self.host_key_checking:
            ssh_options += ' -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no'
        if self.port:
            ssh_options += ' -p {}'.format(self.port)
        return ssh_options

    def __str__(self):
        ssh_command = ''
        if self.password and self.password != ANSIBLE_NULL_VALUE:
            ssh_command += 'sshpass -p "{}"'.format(self.password)

        ssh_command += ' {ssh_exec} {ssh_options} {user_and_hostname}'.format(
            ssh_exec=self.ssh_executable,
            ssh_options=self._get_ssh_options(),
            user_and_hostname=self._get_user_and_hostname()
        )
        return ssh_command


CONNECTION_COMMAND2CLASS_MAP = {
    'ssh': SSHConnectionCommand
}



import argparse
from typing import List, Tuple, no_type_check


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-i',
        '--inventory',
        required=True,
        help='Path to the ansible inventory file'
    )
    parser.add_argument(
        '-g',
        '--groups',
        default=None,
        help="Groups to connect with. Example: -g 'computes:!storage'"
    )
    parser.add_argument(
        '--hosts',
        default=None,
        help="Hostnames to connect with. Example: --hosts 'compute1,storage1'"
    )
    parser.add_argument(
        '-vars',
        '--variables',
        default=None,
        help="Inventory variables to select hosts. Example: -v type:dev,team:ui"
    )
    parser.add_argument(
        '-novars',
        '--no-variables',
        default=None,
        help="Inventory variables to deselect hosts. Example: -novars type:prod,team:sales"
    )
    return parser.parse_args()


def parse_hostnames(hosts: str) -> List[str]:
    """Parse hosts argument into list of hostnames

    :param hosts: String of hosts argument
    :type hosts: str

    :return: List of hostnames
    :rtype: list
    """
    return hosts.split(',') if hosts else []


def parse_inventory_groups(args_groups: str) -> Tuple[List[str], List[str]]:
    """Parse list of inventory groups passed via CLI
    Groups with indices like: 3, [3:], [:3] should be parsed into slices
    that later can be utilizes as list indices on inventory parsing

    :param args_groups: List of strings with groups
    :type args_groups: list

    :return: Two lists of:
                * groups that should be selected
                * groups that should be ommited
    :rtype: list
    """
    if not args_groups:
        return [], []
    provided_groups = args_groups.split(':')
    groups = []
    no_groups = []
    for group in provided_groups:
        if group.startswith('!'):
            no_groups.append(group[1:])
        else:
            groups.append(group)
    return groups, no_groups


@no_type_check
def parse_vars(variables):
    """Parse variables passed as args to list of tuples
    If variable has required value then
    it'll be appended in format (key, value).
    If variable has no variable (it should just exist)
    then it'll be appended as (key,)

    :param variables: string of variables in args format 'key:value,key2:value2,key3'
    :type variables: str

    :return: List of parsed variables
    :rtype: list
    """
    list_of_items = variables.split(',') if variables else []
    parsed_variables = []
    for variable in list_of_items:
        if ':' in variable:
            parsed_variables.append(tuple(variable.split(':')))
        else:
            parsed_variables.append((variable, None))
    return parsed_variables



import pathlib
from setuptools import setup

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()


def requirements(filename):
    with open(filename) as f:
        lines = f.read().splitlines()
    return lines


setup(
    name='ansibleconnect',
    version='1.0.3',
    description='Connect to all hosts from the ansible inventory with one command',
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/psykulsk/ansibleconnect",
    author='Piotr Sykulski',
    author_email='pitersk@gmail.com',
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'Topic :: Utilities',
        'Topic :: System :: Systems Administration'
    ],
    install_requires=requirements('requirements.txt'),
    packages=['ansibleconnect'],
    entry_points='''
    [console_scripts]
    ansibleconnect=ansibleconnect.main:main
    '''
)



#!/usr/bin/env python

import logging
from typing import Iterable

import yaml

from ansibleconnect.ansiblehostadapter import AnsibleHostAdapter
from ansibleconnect.inventoryadapter import InventoryAdapter
from ansibleconnect.parser import parse_arguments, \
    parse_hostnames, \
    parse_inventory_groups, \
    parse_vars
from ansibleconnect.tmuxpresenter import create_tmux_script

logger = logging.getLogger(__name__)


def load_inventory_file(inventory_path: str) -> Iterable:
    with open(inventory_path) as inventory_file:
        inventory_data = yaml.safe_load(inventory_file)
    return inventory_data


def main():
    args = parse_arguments()
    inventory = InventoryAdapter(args.inventory)
    hostnames = parse_hostnames(args.hosts)
    groups, no_groups = parse_inventory_groups(args.groups)
    if hostnames:
        hosts_list = inventory.get_hosts_by_names(hostnames)
    else:
        hosts_list = inventory.get_hosts_by_group(groups, no_groups)
    if not hosts_list:
        print("echo 'No hosts matched given criteria'")
        exit(1)
    variables = parse_vars(args.variables)
    no_variables = parse_vars(args.no_variables)
    if variables or no_variables:
        hosts_list = inventory.get_hosts_by_variables(hosts_list,
                                                      variables,
                                                      no_variables)
    hosts_adapters = [AnsibleHostAdapter(host) for host in hosts_list]
    tmux_script = create_tmux_script(hosts_adapters)
    print(tmux_script)


if __name__ == "__main__":
    main()



from ansible.inventory.host import Host  # type: ignore

from ansibleconnect.connections import CONNECTION_COMMAND2CLASS_MAP
from ansibleconnect.ansible_config_adapter import get_dict_of_ansible_config_options


class AnsibleHostAdapter:
    def __init__(self, ansible_host: Host):
        self._host = ansible_host
        self._connection_plugin = ansible_host.vars.get('ansible_connection', 'ssh')
        self.host_variables = ansible_host.vars
        self.host_variables.update(get_dict_of_ansible_config_options())

    @property
    def connection_command(self):
        return str(
            CONNECTION_COMMAND2CLASS_MAP[self._connection_plugin](self._host.name, self._host.vars))
